.intel_syntax

.section	code,"awx",@progbits

#define		PC	%rsp			/* program counter */
#define		ADDR	%rsi	/* grows up */	/* address stack pointer */
#define		DATA	%rdi	/* grows up */	/* data stack pointer */
#define		TOP	%rax:%rdx		/* top of data stack */

/* scratch %rcx %rbx %r8...%r15 */

.global		bootstrap
.global		exception
.global		shutdown

.global		conputc
.global		congetc
.global		conpeek

.global		character

bootstrap:	enter	0x200,0	/* align the stack */
		add	[pushframe],%rsp
		add	[popframe],%rsp
		call	swapframe
		mov	[stackframe],%rsp
		jmp	exception
shutdown:	mov	%rsp,[stackframe]
		call	swapframe
		leave
		ret

swapframe:	xchg	%rsp,[pushframe]
		push	%r15
		push	%r14
		push	%r13
		push	%r12
		push	%r11
		push	%r10
		push	%r9
		push	%r8
		push	%rdi
		push	%rsi
		push	%rbp
		push	%rsp
		push	%rbx
		push	%rdx
		push	%rcx
		push	%rax
		xchg	%rsp,[popframe]
		pop	%rax
		pop	%rcx
		pop	%rdx
		pop	%rbx
		add	%rsp,8
		pop	%rbp
		pop	%rsi
		pop	%rdi
		pop	%r8
		pop	%r9
		pop	%r10
		pop	%r11
		pop	%r12
		pop	%r13
		pop	%r14
		pop	%r15
		xchg	%rsp,[pushframe]
		ret

.macro		CORO	name
		xchg	%rsp,[stackframe]
		call	swapframe
		call	\name
		call	swapframe
		xchg	%rsp,[stackframe]
.endm

stackframe:	.quad	0

pushframe:	.quad	0x100
popframe:	.quad	0x100

.macro		cell	nn
		.quad	\nn
		.quad	0
.endm

.set		list,0
.macro		WORD	name,len
		.ascii	"\name"
		.byte	\len
		cell	list
.set		list,.
.endm

.macro		NEXT
		ret	8
.endm

exception:	sub	%rax,%rax	/* TOP */
		cqo
		mov	ADDR,offset [addressstack]
		mov	DATA,offset [datastack]
		mov	PC,offset [quit]
		mov	%rbp,offset [swizzle]
		NEXT

quit:		cell	xtQuit
		cell	xtBye

.equ		xtEnter,0xe5ff	/* jmp %rbp */	/*( -- A: xt)*/

swizzle:	add	ADDR,16
		mov	[ADDR],PC
		mov	qword ptr [ADDR+8],0
		mov	PC,[PC-16]
		add	PC,16
		NEXT

		WORD	"exit",4		/*6.1.1380( A: xt --)*/
xtExit:		mov	PC,[ADDR]
		sub	ADDR,16
		NEXT

		WORD	"execute",7		/*6.1.1370( xt --)*/
xtExecute:	add	ADDR,16
		mov	[ADDR],PC
		mov	qword ptr [ADDR+8],0
		mov	PC,offset [trampoline]
		mov	[PC],%rax
		mov	[PC+8],%rdx
		mov	%rax,[DATA]
		mov	%rdx,[DATA+8]
		sub	DATA,16
		NEXT

trampoline:	cell	0
		cell	xtExit

		WORD	"branch",6		/*( PC: xt --)*/
xtBranch:	mov	PC,[PC]
		NEXT

		WORD	"0branch",7		/*( PC: xt D: flag --)*/
xtFalseBranch:	or	%rax,%rdx
		jz	.Lfalsebranch.1
		mov	%rax,[DATA]
		mov	%rdx,[DATA+8]
		sub	DATA,16
		add	PC,16
		NEXT
.Lfalsebranch.1:	mov	%rax,[DATA]
		mov	%rdx,[DATA+8]
		sub	DATA,16
		mov	PC,[PC]
		NEXT

		WORD	"!",1			/*6.1.0010( n a-addr --)*/
xtStore:	mov	%rbx,[DATA]
		mov	%rcx,[DATA+8]
		mov	[%rax],%rbx
		mov	[%rax+8],%rcx
		mov	%rax,[DATA-16]
		mov	%rdx,[DATA-8]
		sub	DATA,32
		NEXT

		WORD	"+",1			/*6.1.0120( n n -- n)*/
xtPlus:		add	%rax,[DATA]
		adc	%rdx,[DATA+8]
		sub	DATA,16
		NEXT

		WORD	"-",1			/*6.1.0160( n n -- n)*/
xtMinus:	mov	%rbx,[DATA]
		mov	%rcx,[DATA+8]
		sub	%rbx,%rax
		sbb	%rcx,%rdx
		mov	%rax,%rbx
		mov	%rdx,%rcx
		sub	DATA,16
		NEXT

		WORD	"0",1			/*( -- n)*/
xtZero:		add	DATA,16
		mov	[DATA],%rax
		mov	[DATA+8],%rdx
		sub	%rax,%rax
		sub	%rdx,%rdx
		NEXT

		WORD	"0<",2			/*6.1.0250( n -- flag)*/
xtZeroLess:	mov	%rbx,0x0000000000000000
		mov	%rcx,0x8000000000000000
		add	%rax,%rbx
		adc	%rdx,%rcx
		sbb	%rax,%rax
		cqo
		NEXT

		WORD	"0<>",3			/*6.2.0260( n -- flag)*/
xtZeroNotEquals:	add	%rax,0xffffffffffffffff
		adc	%rdx,0xffffffffffffffff
		sbb	%rax,%rax
		cqo
		NEXT

		WORD	"0=",2			/*6.1.0270( n -- flag)*/
xtZeroEquals:	sub	%rax,1
		sbb	%rdx,0
		sbb	%rax,%rax
		cqo
		NEXT

		WORD	"0>",2			/*6.2.0280( n -- flag)*/
xtZeroGreater:	sub	%rax,1
		sbb	%rdx,0
		mov	%rbx,0xffffffffffffffff
		mov	%rcx,0x7fffffffffffffff
		sub	%rax,%rbx
		sbb	%rdx,%rcx
		sbb	%rax,%rax
		cqo
		NEXT

		WORD	"1+",2			/*6.1.0290( n -- n)*/
xtInc:		add	%rax,1
		adc	%rdx,0
		NEXT

		WORD	"1-",2			/*6.1.0300( n -- n)*/
xtDec:		sub	%rax,1
		sbb	%rdx,0
		NEXT

		WORD	"2*",2			/*6.1.0320( n -- n)*/
xtTwoStar:	sal	%rax
		rcl	%rdx
		NEXT

		WORD	"2/",2			/*6.1.0330( n -- n)*/
xtTwoSlash:	sar	%rdx
		rcr	%rax
		NEXT

		WORD	"<",1			/*6.1.0480( n n -- flag)*/
xtLessThan:	mov	%rbx,[DATA]
		mov	%rcx,[DATA+8]
		sub	%rbx,%rax
		sbb	%rcx,%rdx
		setl	%al
		add	%al,0xff
		sbb	%rax,%rax
		cqo
		sub	DATA,16
		NEXT

		WORD	">r",2			/*6.1.0580( D: n -- A: n)*/
xtToR:		add	ADDR,16
		mov	[ADDR],%rax
		mov	[ADDR+8],%rdx
		mov	%rax,[DATA]
		mov	%rdx,[DATA+8]
		sub	DATA,16
		NEXT

		WORD	"@",1			/*6.1.0650( a-addr -- n)*/
xtFetch:	mov	%rdx,[%rax+8]
		mov	%rax,[%rax]
		NEXT

		WORD	"aligned",7		/*6.1.0706( c-addr -- a-addr)*/
xtAligned:	add	%rax,15
		adc	%rdx,0
		and	%rax,~15
		and	%rdx,~0
		NEXT

		WORD	"and",3			/*6.1.0720( n n -- n)*/
xtAnd:		and	%rax,[DATA]
		and	%rdx,[DATA+8]
		sub	DATA,16
		NEXT

		WORD	"bye",3			/*15.6.2.0830( --)*/
xtBye:		jmp	shutdown

		WORD	"c!",2			/*6.1.0850( char c-addr --)*/
xtCStore:	movzx	%rcx,byte ptr [DATA]
		mov	[%rax],%cl
		mov	%rax,[DATA-16]
		mov	%rdx,[DATA-8]
		sub	DATA,32
		NEXT

		WORD	"c@",2			/*6.1.0870( c-addr -- char)*/
xtCFetch:	movzx	%rax,byte ptr [%rax]
		mov	%rdx,0
		NEXT

		WORD	"cell+",5		/*6.1.0880( a-addr -- a-addr)*/
xtCellPlus:	add	%rax,16
		adc	%rdx,0
		NEXT

		WORD	"cells",5		/*6.1.0890( n -- n)*/
xtCells:	shld	%rdx,%rax,4
		sal	%rax,4
		NEXT

		WORD	"char+",5		/*6.1.0897( c-addr -- c-addr)*/
xtCharPlus:	add	%rax,1
		adc	%rdx,0
		NEXT

		WORD	"chars",5		/*6.1.0898( n -- n)*/
xtChars:	NEXT

		WORD	"depth",5		/*6.1.1200( -- n)*/
xtDepth:	add	DATA,16
		mov	[DATA],%rax
		mov	[DATA+8],%rdx
		mov	%rax,DATA
		mov	%rdx,0
		sub	%rax,offset [datastack]
		sbb	%rdx,0
		shrd	%rax,%rdx,4
		sar	%rdx,4
		NEXT

		WORD	"drop",4		/*6.1.1260( n --)*/
xtDrop:		mov	%rax,[DATA]
		mov	%rdx,[DATA+8]
		sub	DATA,16
		NEXT

		WORD	"emit",4		/*6.1.1320( char --)*/
xtEmit:		mov	[character],%rax
		mov	[character+8],%rdx
		mov	%rax,[DATA]
		mov	%rdx,[DATA+8]
		sub	DATA,16
		CORO	conputc
		NEXT

character:	cell	0

		WORD	"key",3			/*6.1.1750( -- char)*/
xtKey:		CORO	congetc
		add	DATA,16
		mov	[DATA],%rax
		mov	[DATA+8],%rdx
		movsx	%rax,word ptr [character]
		cqo
		NEXT

		WORD	"key?",4		/*10.6.1.1755( -- flag)*/
xtKeyQuestion:	CORO	conpeek
		add	DATA,16
		mov	[DATA],%rax
		mov	[DATA+8],%rdx
		cmp	word ptr [character],0x8000
		sbb	%rax,%rax
		cqo
		NEXT

		WORD	"literal",7		/*6.1.1780( PC: n -- D: n)*/
doesLiteral:	add	DATA,16
		mov	[DATA],%rax
		mov	[DATA+8],%rdx
		pop	%rax
		pop	%rdx
		NEXT

		WORD	"lshift",6		/*6.1.1805( n n -- n)*/
xtLShift:	cmp	%rdx,0
		jl	exception
		mov	%rcx,%rax
		mov	%rax,[DATA]
		mov	%rdx,[DATA+8]
		cmp	%rcx,64
		jb	.Llshift.1
		mov	%rdx,%rax
		mov	%rax,0
		sub	%rcx,64
		cmp	%rcx,64
		jb	.Llshift.1
		mov	%rdx,0
		mov	%cl,0
.Llshift.1:	shld	%rdx,%rax,%cl
		shl	%rax,%cl
		sub	DATA,16
		NEXT

		WORD	"not",3			/*( n -- n)*/
xtNot:		not	%rax
		not	%rdx
		NEXT

		WORD	"or",2			/*6.1.1980( n n -- n)*/
xtOr:		or	%rax,[DATA]
		or	%rdx,[DATA+8]
		sub	DATA,16
		NEXT

		WORD	"pick",4		/*6.2.2030( 1 2 n -- 1 2 1)*/
xtPick:		cmp	%rdx,0
		jl	exception
		sal	%rax,4
		mov	%rbx,DATA
		sub	%rbx,%rax
		mov	%rax,[%rbx]
		mov	%rdx,[%rbx+8]
		NEXT

		WORD	"r-depth",7		/*( -- n)*/
xtRDepth:	add	DATA,16
		mov	[DATA],%rax
		mov	[DATA+8],%rdx
		mov	%rax,ADDR
		mov	%rdx,0
		sub	%rax,offset [addressstack]
		sbb	%rdx,0
		shrd	%rax,%rdx,4
		sar	%rdx,4
		NEXT

		WORD	"r-drop",6		/*( A: n --)*/
xtRDrop:	sub	ADDR,16
		NEXT

		WORD	"r-pick",6		/*( A: 1 2 D: n -- A: 1 2 1)*/
xtRPick:	cmp	%rdx,0
		jl	exception
		sal	%rax,4
		mov	%rbx,ADDR
		sub	%rbx,%rax
		mov	%rax,[%rbx]
		mov	%rdx,[%rbx+8]
		add	ADDR,16
		mov	[ADDR],%rax
		mov	[ADDR+8],%rdx
		mov	%rax,[DATA]
		mov	%rdx,[DATA+8]
		sub	DATA,16
		NEXT

		WORD	"r-roll",6		/*( A: 1 2 3 D: n -- A: 2 3 1)*/
xtRRoll:	cmp	%rdx,0
		jl	exception
		mov	%rcx,%rax
		jrcxz	.Lrroll.2
		mov	%rax,[ADDR]
		mov	%rdx,[ADDR+8]
		mov	%rbx,ADDR
.Lrroll.1:	sub	%rbx,16
		mov	%r8,[%rbx]
		mov	%r9,[%rbx+8]
		mov	[%rbx],%rax
		mov	[%rbx+8],%rdx
		mov	%rax,%r8
		mov	%rdx,%r9
		loop	.Lrroll.1
		mov	[ADDR],%rax
		mov	[ADDR+8],%rdx
.Lrroll.2:	mov	%rax,[DATA]
		mov	%rdx,[DATA+8]
		sub	DATA,16
		NEXT

		WORD	"r>",2			/*6.1.2060( A: n -- D: n)*/
xtRFrom:	add	DATA,16
		mov	[DATA],%rax
		mov	[DATA+8],%rdx
		mov	%rax,[ADDR]
		mov	%rdx,[ADDR+8]
		sub	ADDR,16
		NEXT

		WORD	"roll",4		/*6.2.2150( 1 2 3 n -- 2 3 1)*/
xtRoll:		cmp	%rdx,0
		jl	exception
		mov	%rcx,%rax
		mov	%rax,[DATA]
		mov	%rdx,[DATA+8]
		jrcxz	.Lroll.2
		mov	%rbx,DATA
.Lroll.1:	sub	%rbx,16
		mov	%r8,[%rbx]
		mov	%r9,[%rbx+8]
		mov	[%rbx],%rax
		mov	[%rbx+8],%rdx
		mov	%rax,%r8
		mov	%rdx,%r9
		loop	.Lroll.1
.Lroll.2:	sub	DATA,16
		NEXT

		WORD	"rshift",6		/*6.1.2162( n n -- n)*/
xtRShift:	cmp	%rdx,0
		jl	exception
		mov	%rcx,%rax
		mov	%rax,[DATA]
		mov	%rdx,[DATA+8]
		cmp	%rcx,64
		jb	.Lrshift.1
		mov	%rax,%rdx
		mov	%rdx,0
		sub	%rcx,64
		cmp	%rcx,64
		jb	.Lrshift.1
		mov	%rax,0
		mov	%cl,0
.Lrshift.1:	shrd	%rax,%rdx,%cl
		shr	%rdx,%cl
		sub	DATA,16
		NEXT

		WORD	"u<",2			/*6.1.2340( n n -- flag)*/
xtULessThan:	mov	%rbx,[DATA]
		mov	%rcx,[DATA+8]
		sub	%rbx,%rax
		sbb	%rcx,%rdx
		setb	%al
		add	%al,0xff
		sbb	%rax,%rax
		cqo
		sub	DATA,16
		NEXT

		WORD	"um+",3			/*( d n -- d)*/
xtUMPlus:	add	[DATA-16],%rax
		adc	[DATA-8],%rdx
		adc	qword ptr [DATA],0
		adc	qword ptr [DATA+8],0
		mov	%rax,[DATA]
		mov	%rdx,[DATA+8]
		sub	DATA,16
		NEXT

		WORD	"um-",3			/*( d n -- d)*/
xtUMMinus:	sub	[DATA-16],%rax
		sbb	[DATA-8],%rdx
		sbb	qword ptr [DATA],0
		sbb	qword ptr [DATA+8],0
		mov	%rax,[DATA]
		mov	%rdx,[DATA+8]
		sub	DATA,16
		NEXT

.struct		0
Iproductlo:	.space	16
Iproducthi:	.space	16
Imultiplicand:	.space	16
Imultiplier:	.space	16
Instar:		.byte	0
.previous

		WORD	"um*",3			/*6.1.2360( n n -- d)*/
xtUMStar:	mov	[DATA+Imultiplier],%rax
		mov	[DATA+Imultiplier+8],%rdx
		mov	%rax,[DATA]
		mov	%rdx,[DATA+8]
		mov	[DATA+Imultiplicand],%rax
		mov	[DATA+Imultiplicand+8],%rdx

		mov	%rax,0
		mov	%rbx,0
		mov	%rcx,0
		mov	%rdx,0
		mov	%r8,0
		mov	%r9,0
		mov	%r10,[DATA+Imultiplicand]
		mov	%r11,[DATA+Imultiplicand+8]
		mov	byte ptr [DATA+Instar],128
.Lumstar.1:	shr	%r11
		rcr	%r10
		rcr	%r9
		rcr	%r8
		shl	qword ptr [DATA+Imultiplier]
		rcl	qword ptr [DATA+Imultiplier+8]
		jnc	.Lumstar.2
		add	%rax,%r8
		adc	%rbx,%r9
		adc	%rcx,%r10
		adc	%rdx,%r11
.Lumstar.2:	dec	byte ptr [DATA+Instar]
		jnz	.Lumstar.1
		mov	[DATA+Iproductlo],%rax
		mov	[DATA+Iproductlo+8],%rbx
		mov	[DATA+Iproducthi],%rcx
		mov	[DATA+Iproducthi+8],%rdx

		mov	%rax,[DATA+Iproducthi]
		mov	%rdx,[DATA+Iproducthi+8]
		NEXT

.struct		0
Iremainder:
Idividendlo:	.space	16
Idividendhi:	.space	16
Idivisor:	.space	16
Iquotient:	.space	16
Inslashmod:	.byte	0
.previous

		WORD	"um/mod",6		/*6.1.2370( d n -- n n)*/
xtUMSlashMod:	mov	%rbx,[DATA]
		mov	%rcx,[DATA+8]
		sub	%rbx,%rax
		sbb	%rcx,%rdx
		jnc	exception
		sub	DATA,16
		mov	[DATA+Idivisor],%rax
		mov	[DATA+Idivisor+8],%rdx

		mov	%rax,[DATA+Idividendlo]
		mov	%rbx,[DATA+Idividendlo+8]
		mov	%rcx,[DATA+Idividendhi]
		mov	%rdx,[DATA+Idividendhi+8]
		mov	%r8,0
		mov	%r9,0
		mov	%r10,[DATA+Idivisor]
		mov	%r11,[DATA+Idivisor+8]
		mov	qword ptr [DATA+Iquotient],0
		mov	qword ptr [DATA+Iquotient+8],0
		mov	byte ptr [DATA+Inslashmod],128
.Lumslashmod.1:	shr	%r11
		rcr	%r10
		rcr	%r9
		rcr	%r8
		sub	%rax,%r8
		sbb	%rbx,%r9
		sbb	%rcx,%r10
		sbb	%rdx,%r11
		jnc	.Lumslashmod.2
		add	%rax,%r8
		adc	%rbx,%r9
		adc	%rcx,%r10
		adc	%rdx,%r11
.Lumslashmod.2:	rcl	qword ptr [DATA+Iquotient]
		rcl	qword ptr [DATA+Iquotient+8]
		dec	byte ptr [DATA+Inslashmod]
		jnz	.Lumslashmod.1
		not	qword ptr [DATA+Iquotient]
		not	qword ptr [DATA+Iquotient+8]
		mov	[DATA+Iremainder],%rax
		mov	[DATA+Iremainder+8],%rbx

		mov	%rax,[DATA+Iquotient]
		mov	%rdx,[DATA+Iquotient+8]
		NEXT

.struct		0
Iremainderlo:
Iradicandlo:	.space	16
Iremainderhi:
Iradicandhi:	.space	16
Iroot:		.space	16
Insqrt:		.byte	0
.previous

		WORD	"umsqrt",6		/*( d -- n)*/
xtUMSqrt:	mov	[DATA+Iradicandhi],%rax
		mov	[DATA+Iradicandhi+8],%rdx

		mov	%rax,0
		mov	%rbx,0
		mov	%rcx,0
		mov	%rdx,0
		mov	%r8,0
		mov	%r9,0
		mov	%r10,0
		mov	%r11,0x8000000000000000
		mov	byte ptr [DATA+Insqrt],128
.Lumsqrt.1:	shr	%r11
		rcr	%r10
		rcr	%r9
		rcr	%r8
		add	%rax,%r8
		adc	%rbx,%r9
		adc	%rcx,%r10
		adc	%rdx,%r11
		sub	[DATA+Iradicandlo],%rax
		sbb	[DATA+Iradicandlo+8],%rbx
		sbb	[DATA+Iradicandhi],%rcx
		sbb	[DATA+Iradicandhi+8],%rdx
		jnc	.Lumsqrt.2
		add	[DATA+Iradicandlo],%rax
		adc	[DATA+Iradicandlo+8],%rbx
		adc	[DATA+Iradicandhi],%rcx
		adc	[DATA+Iradicandhi+8],%rdx
		sub	%rax,%r8
		sbb	%rbx,%r9
		sbb	%rcx,%r10
		sbb	%rdx,%r11
		jmp	.Lumsqrt.3
.Lumsqrt.2:	add	%rax,%r8
		adc	%rbx,%r9
		adc	%rcx,%r10
		adc	%rdx,%r11
.Lumsqrt.3:	shr	%rdx
		rcr	%rcx
		rcr	%rbx
		rcr	%rax
		shr	%r11
		rcr	%r10
		rcr	%r9
		rcr	%r8
		dec	byte ptr [DATA+Insqrt]
		jnz	.Lumsqrt.1
		mov	[DATA+Iroot],%rax
		mov	[DATA+Iroot+8],%rbx

		mov	%rax,[DATA+Iroot]
		mov	%rdx,[DATA+Iroot+8]
		sub	DATA,16
		NEXT

.section	data,"awx",@progbits

#include	"threaded.S"

.section	bss,"awx",@nobits

padspace:	.space	4096
addressstack:	.space	64*1024
datastack:	.space	64*1024
dataspace:	.space	1*1024*1024

.end
