/* global variables visible to Forth and the Host */

extern void bootstrap(void);
extern void exception(void);
extern void shutdown(void);

extern void conputc(void);
extern void congetc(void);
extern void conpeek(void);

extern int character;

/* POSIX operating system interface */

/* extremely machine dependent */

#ifdef __NetBSD__
#include <signal.h>
#endif
#ifdef __FreeBSD__
#include <signal.h>
#endif
#ifdef __linux__
#define _GNU_SOURCE
#include <signal.h>
#endif

void handler(int signo, siginfo_t *info, void *context)
{
	ucontext_t *uc = context;
#ifdef __NetBSD__
	_UC_MACHINE_SET_PC(uc, exception);
#endif
#ifdef __FreeBSD__
#ifdef __i386
	uc->uc_mcontext.mc_eip = exception;
#endif
#ifdef __amd64
	uc->uc_mcontext.mc_rip = exception;
#endif
#endif
#ifdef __linux__
#ifdef __i386
	uc->uc_mcontext.gregs[REG_EIP] = exception;
#endif
#ifdef __amd64
	uc->uc_mcontext.gregs[REG_RIP] = exception;
#endif
#endif
}

stack_t ss;
struct sigaction sa;

#include <termios.h>

struct termios tt;
struct termios ot;

#include <stdio.h>
#include <unistd.h>

int cc = EOF; /* buffer */

void conputc(void)
{
	fputc(character, stdout);
}
void congetc(void)
{
	if(cc == EOF) {
		if(tt.c_cc[VMIN] != 1) {
			tt.c_cc[VMIN] = 1;
			if(tcsetattr(STDIN_FILENO, TCSANOW, &tt) == -1);
		}
		cc = fgetc(stdin);
	}
	character = cc;
	cc = EOF;
}
void conpeek(void)
{
	if(cc == EOF) {
		if(tt.c_cc[VMIN] != 0) {
			tt.c_cc[VMIN] = 0;
			if(tcsetattr(STDIN_FILENO, TCSANOW, &tt) == -1);
		}
		cc = fgetc(stdin);
	}
	character = cc;
}

#include <stdlib.h>

void quit(void)
{
	if(tcsetattr(STDIN_FILENO, TCSANOW, &ot) == -1) perror("tcsetattr");
}
int main(void)
{
	char sp[SIGSTKSZ]; /* alternate stack */
	ss.ss_sp = sp;
	ss.ss_size = sizeof(sp);
	ss.ss_flags = 0;
	if(sigaltstack(&ss, 0) == -1) perror("sigaltstack");
	sa.sa_sigaction = handler;
	sa.sa_flags = SA_ONSTACK | SA_RESETHAND | SA_NODEFER | SA_SIGINFO;
	if(sigemptyset(&sa.sa_mask) == -1) perror("sigemptyset");
	if(sigaction(SIGILL, &sa, 0) == -1) perror("sigaction sigill");
	if(sigaction(SIGFPE, &sa, 0) == -1) perror("sigaction sigfpe");
	if(sigaction(SIGBUS, &sa, 0) == -1) perror("sigaction sigbus");
	if(sigaction(SIGSEGV, &sa, 0) == -1) perror("sigaction sigsegv");
	if(tcgetattr(STDIN_FILENO, &ot) == -1) perror("tcgetattr");
	if(atexit(quit) == -1) perror("atexit");
	tt = ot;
	tt.c_lflag &= ~(ECHO | ICANON);
	tt.c_cc[VMIN] = 1;
	tt.c_cc[VTIME] = 0;
	if(tcsetattr(STDIN_FILENO, TCSANOW, &tt) == -1) perror("tcsetattr");
	bootstrap();
	exit(0);
	return(0);
}
